#include "system_monitor.hpp"

using namespace std;

system_monitor::system_monitor(DDSDomainParticipant *participant) : participant_ (participant)
{

    //Getting QoS files
    string sysconf_dir(SYSCONFDIR);    
    qos_file_ = sysconf_dir + "qos/system_monitor_publisher_qos.xml";


	//Temperature Stuff
	list <string> therms;
	string path;
	sensor_type_ = SENSOR_TYPE_NONE;
	sensor_name_ = NULL;
	sensor_path_ = NULL;
	sensor_alarm_value_ = 60;

	if ((!sensor_type_) || ((!sensor_name_) || (sensor_name_[0] == 0))) {
		if (sensor_name_) free(sensor_name_);
		if (sensor_path_) free(sensor_path_);
		sensor_path_ = NULL;

		therms = fm_.ls("/proc/acpi/thermal_zone");
		if (!therms.empty()) {
			string name;
			name = therms.back();
			sensor_type_ = SENSOR_TYPE_LINUX_ACPI;
			sensor_name_ = strdup(name.c_str());

			therms.clear();
		} else {
			therms.clear();
			if (fm_.exists("/sys/devices/virtual/thermal")) {
				sensor_type_ = SENSOR_TYPE_LINUX_VIRTUAL;
				sensor_name_ = strdup("dummy");
			} else if (fm_.exists("/proc/omnibook/temperature")) {
				sensor_type_ = SENSOR_TYPE_OMNIBOOK;
				sensor_name_ = strdup("dummy");
			} else if (fm_.exists("/sys/devices/temperatures/sensor1_temperature")) {
				sensor_type_ = SENSOR_TYPE_LINUX_PBOOK;
				sensor_name_ = strdup("dummy");
			} else if (fm_.exists("/sys/devices/temperatures/cpu_temperature")) {
				sensor_type_ = SENSOR_TYPE_LINUX_MACMINI;
				sensor_name_ = strdup("dummy");
			} else if (fm_.exists("/sys/devices/platform/coretemp.0/temp1_input")) {
				sensor_type_ = SENSOR_TYPE_LINUX_INTELCORETEMP;
				sensor_name_ = strdup("dummy");
			}                ///*
			else {
				// try the i2c bus

				therms = temperature_get_bus_files("i2c");
				if (!therms.empty()) {
					string name;
					name = therms.back();
					if (fm_.exists(name)) {
						path = fm_.file_get(name);
						sensor_type_ = SENSOR_TYPE_LINUX_I2C;
						sensor_path_ = strdup(name.c_str());
						sensor_name_ = strdup(path.c_str());
					}
					therms.clear();
				}
				if (!sensor_path_) {
					// try the pci bus
					therms = temperature_get_bus_files("pci");
					if (!therms.empty()) {
						string name;

						name = therms.back();

						if (fm_.exists(name)) {
							path = fm_.file_get(name);
							sensor_type_ = SENSOR_TYPE_LINUX_PCI;
							sensor_path_ = strdup(name.c_str());
							sensor_name_ = strdup(path.c_str());
						}

						therms.clear();
					}
				}
			}
		}

	}
	if ((sensor_type_) && (sensor_name_) && (!sensor_path_)) {
		switch (sensor_type_) {
		case SENSOR_TYPE_NONE:
			break;
		case SENSOR_TYPE_OMNIBOOK:
			sensor_path_ = strdup("/proc/omnibook/temperature");
			break;
		case SENSOR_TYPE_LINUX_MACMINI:
			sensor_path_ = strdup("/sys/devices/temperatures/cpu_temperature");
			break;
		case SENSOR_TYPE_LINUX_PBOOK:
			sensor_path_ = strdup("/sys/devices/temperatures/sensor1_temperature");
			break;
		case SENSOR_TYPE_LINUX_INTELCORETEMP:
			sensor_path_ = strdup("/sys/devices/platform/coretemp.0/temp1_input");
			break;
		case SENSOR_TYPE_LINUX_I2C:
			therms = fm_.ls("/sys/bus/i2c/devices");

			for (list<string>::iterator it = therms.begin(); it != therms.end(); ++it) {
				path = "/sys/bus/i2c/devices/" + *it + "/" + sensor_name_ + "_input";
				if (fm_.exists(path)) {
					sensor_path_ = strdup(path.c_str());
					/* We really only care about the first
					 * one for the default. */
					break;
				}
			}
			break;
		case SENSOR_TYPE_LINUX_PCI:
			therms = fm_.ls("/sys/bus/pci/devices");
			for (list<string>::iterator it = therms.begin(); it != therms.end(); ++it) {
				path = "/sys/bus/pci/devices/" + *it + "/" + sensor_name_ + "_input";
				if (fm_.exists(path)) {
					sensor_path_ = strdup(path.c_str());
					/* We really only care about the first
					 * one for the default. */
					break;
				}
			}

			break;

		case SENSOR_TYPE_LINUX_ACPI:
			path = "/proc/acpi/thermal_zone/" + string(sensor_name_) + "/temperature";
			sensor_path_ = strdup(path.c_str());
			break;
		case SENSOR_TYPE_LINUX_VIRTUAL:
			string busdir = "/sys/devices/virtual/thermal";
			therms = fm_.ls(busdir);
			for (list<string>::iterator name = therms.begin(); name != therms.end(); ++name) {
				path = busdir + "/" + *name + "/" + "temp";
				if (fm_.exists(path)) {
					sensor_path_ = strdup(path.c_str());
					/* We really only care about the first
					 * one for the default. */
					break;

				}
			}
			break;
		}
	}



	/*DDS Stuff*/

	/* To customize participant QoS, use
	 * the configuration file USER_QOS_PROFILES.xml */
	instance_handle_ = DDS_HANDLE_NIL;

	/* To customize publisher QoS, use
	 * the configuration file USER_QOS_PROFILES.xml */
	publisher_ = participant_->create_publisher(
			DDS_PUBLISHER_QOS_DEFAULT, NULL /* listener */, DDS_STATUS_MASK_NONE);

	if (publisher_ == NULL) {
        syslog(LOG_ERR,"SystemMonitor plugin: create_publisher error");
	}

	/* Register type before creating topic */
	type_name_ = sysmon_statusTypeSupport::get_type_name();
	retcode_ = sysmon_statusTypeSupport::register_type(
			participant_, type_name_);
	if (retcode_ != DDS_RETCODE_OK) {
        syslog(LOG_ERR,"SystemMonitor plugin: register_type error %d", retcode_);
	}


	/* To customize topic QoS, use
       the configuration file USER_QOS_PROFILES.xml */
	topic_ = participant->create_topic(
			"sysmon_status",
			type_name_, DDS_TOPIC_QOS_DEFAULT, NULL /* listener */,
			DDS_STATUS_MASK_NONE);
	if (topic_ == NULL) {
         syslog(LOG_ERR,"SystemMonitor plugin: create_topic error");
	}

	/* To customize data writer QoS, use
       the configuration file USER_QOS_PROFILES.xml */
	writer_ = publisher_->create_datawriter(
			topic_, DDS_DATAWRITER_QOS_DEFAULT, NULL /* listener */,
			DDS_STATUS_MASK_NONE);
	if (writer_ == NULL) {
		 syslog(LOG_ERR,"SystemMonitor plugin: create_datawriter error");
	}
	sysmon_status_writer_ = sysmon_statusDataWriter::narrow(writer_);
	if (sysmon_status_writer_ == NULL) {
		 syslog(LOG_ERR,"SystemMonitor plugin: DataWriter narrow error");
	}

	/* Create data sample for writing */
	instance_ = sysmon_statusTypeSupport::create_data();
	if (instance_ == NULL) {
		 syslog(LOG_ERR,"SystemMonitor plugin: sysmon_statusTypeSupport::create_data error");
	}

    glibtop_init();
    static struct utsname name; 
	uname(&name);            

	hi_.hostname = name.nodename;
	hi_.sysname = name.sysname;
	hi_.release_name =  name.release;
	hi_.machine = name.machine; 	

}

list <string> system_monitor::temperature_get_bus_files(string bus) {

    string path, busdir;
    list <string> result, therms;

    busdir = "/sys/bus/" + bus + "/devices";

    therms = fm_.ls(busdir);


    for (list<string>::iterator name = therms.begin(); name != therms.end(); ++name) {

        list<string> files;
        path = busdir + "/" + *name;
        files = fm_.ls(path);

        for (list<string>::iterator file = files.begin(); file != files.end(); ++file) {

            if ((!strncmp("temp", (*file).c_str(), 4)) &&
                    (!strcmp("_input", &((*file).c_str())[strlen((*file).c_str()) - 6]))) {

                char *f;
                path = busdir + "/" + *name + "/" + *file;

                f = strdup(path.c_str());
                if (f)
                    result.push_back(string(f));
            }
        }
    }

    return result;

}

double system_monitor::get_temperature(void)
{
	FILE *f;

	bool error = false;

	int ret = 0;
	double temp = 0.0;
	char buf[PATH_MAX];

	switch (sensor_type_) {
	case SENSOR_TYPE_NONE:
		break;

	case SENSOR_TYPE_OMNIBOOK:
		f = fopen(sensor_path_, "rb");

		if (f) {
			char dummy[4096];

			fgets(buf, sizeof (buf), f);
			buf[sizeof (buf) - 1] = 0;
			if (sscanf(buf, "%s %s %f", dummy, dummy, &temp) == 3)
				ret = 1;
			else {
				error = true;
				break;
			}
			fclose(f);
		} else
			error = true;

		break;
	case SENSOR_TYPE_LINUX_MACMINI:
	case SENSOR_TYPE_LINUX_PBOOK:
		f = fopen(sensor_path_, "rb");
		if (f) {
			fgets(buf, sizeof (buf), f);
			buf[sizeof (buf) - 1] = 0;
			if (sscanf(buf, "%f", &temp) == 1)
				ret = 1;
			else {
				error = true;
				break;
			}
			fclose(f);
		} else
			error = true;
		break;
	case SENSOR_TYPE_LINUX_VIRTUAL:
	case SENSOR_TYPE_LINUX_INTELCORETEMP:
	case SENSOR_TYPE_LINUX_I2C:
		f = fopen(sensor_path_, "r");
		if (f) {
			fgets(buf, sizeof (buf), f);
			buf[sizeof (buf) - 1] = 0;

			/* actually read the temp */
			if (sscanf(buf, "%f", &temp) == 1)
				ret = 1;
			else {
				error = true;
				break;
			}
			temp = temp / 1000.;
			fclose(f);
		} else
			error = true;
		break;
	case SENSOR_TYPE_LINUX_PCI:
		f = fopen(sensor_path_, "r");
		if (f) {
			fgets(buf, sizeof (buf), f);
			buf[sizeof (buf) - 1] = 0;

			/* actually read the temp */
			if (sscanf(buf, "%f", &temp) == 1)
				ret = 1;
			else {
				error = true;
				break;
			}
			/* Hack for temp */
			temp = temp / 1000.;
			fclose(f);
		} else
			error = true;
		break;
	case SENSOR_TYPE_LINUX_ACPI:
		f = fopen(sensor_path_, "r");
		if (f) {
			char *p, *q;

			fgets(buf, sizeof (buf), f);
			buf[sizeof (buf) - 1] = 0;
			fclose(f);
			p = strchr(buf, ':');
			if (p) {
				p++;
				while (*p == ' ') p++;
				q = strchr(p, ' ');
				if (q) *q = 0;
				temp = atoi(p);
				ret = 1;
			} else
				error = true;
		} else
			error = true;
		break;
	}
	if (ret) {
		return temp;
	}
	else if(error) {
		sensor_type_ = SENSOR_TYPE_NONE;
		if (sensor_name_) free(sensor_name_);
		sensor_name_ = NULL;
		if (sensor_path_) free(sensor_path_);
		sensor_path_ = NULL;
	}
	return -1.0;
}


void system_monitor::publish_information()
{
	//sysmon_status sm;
    instance_->host = hi_;
	//Temperature (ºC)
	instance_->cpu_temp = get_temperature();
	//CPU (%)
	glibtop_get_cpu (&cpu_);
	instance_->cpu_user = cpu_.user*100./cpu_.total;
	instance_->cpu_system = cpu_.sys*100./cpu_.total;
	instance_->cpu_idle = cpu_.idle*100./cpu_.total;
	//Load average
	glibtop_get_loadavg(&loadavg_);
	instance_->lavg.load_one = loadavg_.loadavg [0];
	instance_->lavg.load_five = loadavg_.loadavg [1];
	instance_->lavg.load_fifteen = loadavg_.loadavg [2];
	//Memory (KB)
	glibtop_get_mem(&memory_);
	instance_->mem_total = memory_.total/(1024.0);
	instance_->mem_used = memory_.used/(1024.0);
	instance_->mem_free = memory_.free/(1024.0);
	//Swap (KB)
	glibtop_get_swap(&swap_);
	instance_->swap_total = swap_.total/(1024.0);
	instance_->swap_used = swap_.used/(1024.0);
	instance_->swap_free = swap_.free/(1024.0);

	//Publish DDS
	retcode_ = sysmon_status_writer_->write(*instance_, instance_handle_);
	if (retcode_ != DDS_RETCODE_OK) {
		 syslog(LOG_ERR,"SystemMonitor plugin: write error %d", retcode_);
	}


//	cout << "Hostname" << endl;
//	cout << sm.hostname << endl << endl;
//
//	cout << "Temperature" << endl;
//	cout << sm.cpu_temp << endl;
//	cout << endl;
//
//	cout << "CPU" << endl;
//	cout << "User " << sm.cpu_user << endl;
//	cout << "Sys " << sm.cpu_system << endl;
//	cout << "Idle " << sm.cpu_idle << endl;
//	cout << endl;
//
//	cout << "Load average" << endl;
//	cout << sm.lavg.load_one << endl;
//	cout << sm.lavg.load_five << endl;
//	cout << sm.lavg.load_fifteen << endl;
//	cout << endl;
//
//	cout << "Memory" << endl;
//	cout << sm.mem_total << "MB" << endl;
//	cout << sm.mem_used << "MB" << endl;
//	cout << sm.mem_free << "MB" << endl;
//	cout << endl;
//
//	cout << "Swap" << endl;
//	cout << sm.swap_total<< "MB" << endl;
//	cout << sm.swap_used << "MB"<< endl;
//	cout << sm.swap_free<< "MB" << endl;
//	cout << endl;
}


system_monitor::~system_monitor(void)
{
	glibtop_close();
}

//file_management class

bool file_management::is_dir(string file)
{
	struct stat st;

	if (stat(file.c_str(), &st) < 0) return false;
	if (S_ISDIR(st.st_mode)) return true;
	return false;
}

string file_management::get_realpath(string file)
{
	char buf[PATH_MAX];
	if (!file.c_str()) return string(strdup(""));
	if (!realpath(file.c_str(), buf)) return string(strdup(""));
	return string(strdup(buf));
}

bool file_management::mkidr(string dir)
{
	if (mkdir(dir.c_str(), default_mode) < 0) return false;
	return true;
}

bool file_management::rmmdir(string dir)
{
	if (rmdir(dir.c_str()) < 0) return false;
	return true;
}

string file_management::dir_get(string file)
{
	char *p;
	char buf[PATH_MAX];
	strncpy(buf, file.c_str(), PATH_MAX);
	p = strrchr(buf, '/');
	if (!p) {
		return strdup(file.c_str());
	}

	if (p == buf) return strdup("/");

	*p = 0;
	return string(strdup(buf));
}

string file_management::file_get(string str)
{
	size_t found;
	found = str.find_last_of("/\\");
	return str.substr(found + 1);
}

bool file_management::has_extension(string file, string extension)
{
	size_t found;
	found = file.find_last_of(".");
	if (file.substr(found + 1) == extension)
		return true;
	return false;
}

list<string> file_management::ls_r_by_extension(string dir, string extension)
{

	std::list<string> file_list;
	DIR *dirp = opendir(dir.c_str());

	if (dirp) {
		struct dirent *dp = NULL;

		while ((dp = readdir(dirp)) != NULL) {
			std::string file(dp->d_name);

			if (file == "." || file == "..") // skip these
				continue;

			if (dp->d_type & DT_DIR) {
				// found a directory; recurse into it.
				string file_path = dir + "/" + file;
				list <string> aux = ls_r_by_extension(file_path, extension);
				file_list.insert(file_list.end(), aux.begin(), aux.end());
			} else {
				if (has_extension(file, extension)) {
					file_list.push_back(dir + "/" + file);
				}
			}
		}

		closedir(dirp);
	}

	return file_list;

}

list <string> file_management::ls(string dir)
{
	string f;
	DIR *dirp;
	list <string> files;
	struct dirent *dp;

	dirp = opendir(dir.c_str());
	if (!dirp) return files;

	while ((dp = readdir(dirp))) {
		if ((strcmp(dp->d_name, ".")) && (strcmp(dp->d_name, ".."))) {
			f = string(strdup(dp->d_name));
			files.push_back(f);
		}
	}

	closedir(dirp);

	return files;

}

bool file_management::exists(string file)
{
	struct stat st;

	if (stat(file.c_str(), &st) < 0 && file != "/")
		return false;

	return true;

}


